/**
 * @ingroup   emu68_devel
 * @file      emu68.c
 * @author    Ben(jamin) Gerard <ben@sashipa.com>
 * @date      1998/06/10
 * @brief     Condition code function table.
 * @version   $Id$
 */

/* Copyright (C) 1998-2001 Ben(jamin) Gerard */

#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "emu68/emu68.h"

/*************************************************************
* Pointer to unique interruping IO (MFP for ST emulation!!!) *
*************************************************************/
io68_t *interrupt_io = NULL;

/***************************
* 68000 internal registers *
***************************/
reg68_t reg68;

extern linefunc68_t *line_func[1024];

/*******************************
* EMU68 internal struct access *
*******************************/

/*  Set new interrupt IO
 *  return previous pointer
*/
io68_t *EMU68_set_interrupt_io( io68_t *io )
{
  io68_t *oldio = interrupt_io;
  interrupt_io = io;
  return oldio;
}

void EMU68_set_registers(reg68_t *r)
{
  int i;
  if(r==NULL) return;
  reg68.usp = r->usp;
  reg68.pc  = r->pc;
  reg68.sr  = r->sr;
  for(i=0; i<16; i++)
    reg68.d[i] = r->d[i];
}

void EMU68_get_registers(reg68_t *r)
{
  int i;
  if(r==NULL) return;
  r->usp = reg68.usp;
  r->pc  = reg68.pc;
  r->sr  = reg68.sr;
  for(i=0; i<16; i++)
    r->d[i] = reg68.d[i];
}

void EMU68_set_cycle(u32 cycle)
{
  reg68.cycle = cycle;
}

u32 EMU68_get_cycle(void)
{
  return reg68.cycle;
}

/*******************************
* EMU68 on board memory access *
*******************************/
int EMU68_memvalid(u32 dest, u32 sz)
{
  if(sz>reg68.memsz)
    return EMU68error_add("Not enought 68K memory ($%X>$%X)",sz,reg68.memsz);
  sz = (dest+sz)&reg68.memmsk;
  if(sz<dest)
    return EMU68error_add("68K memory overflow :($%X-$%X)",dest,sz);
  return 0;
}

/* Peek & Poke
*/
u8 EMU68_peek(u32 addr)
{
  return reg68.mem[addr&reg68.memmsk];
}

u8 EMU68_poke(u32 addr, u8 v)
{
  return reg68.mem[addr&reg68.memmsk] = v;
}

/*  Write memory block to 68K on board memory
*/
int EMU68_memput(u32 dest, u8 *src, u32 sz)
{
  int err;
  if(err=EMU68_memvalid(dest,sz), err<0)
    return err;
  memcpy(reg68.mem+dest,src,sz);
  return 0;
}

/*  Read memory block from 68K on board memory
*/
int EMU68_memget(u8 *dest, u32 src, u32 sz)
{
  int err;
  if(err=EMU68_memvalid(src,sz), err<0)
    return err;
  memcpy(dest,reg68.mem+src,sz);
  return 0;
}

/* 68000 Hardware Reset
 * - PC = 0
 * - SR = 2700
 * - A7 = end of mem - 4
 * - All registers cleared
 * - All IO reseted
*/
void EMU68_reset(void)
{
  io68_t *io;
  for(io=reg68.iohead; io!=NULL; io=io->next)
    (io->reset)();
  memset(reg68.d,0,sizeof(reg68.d)*2);
  reg68.cycle  = 0;
  reg68.pc     = 0;
  reg68.sr     = 0x2700;
  reg68.a[7]   = reg68.memsz-4;
  reg68.status = 0;

}

/* 68000 OP-WORD format :
 * 1111 0000 0000 0000 ( LINE  )
 * 0000 0001 1111 1000 ( OP    )
 * 0000 0000 0000 0111 ( REG 0 )
 * 0000 1110 0000 0000 ( REG 9 )
*/

/* Emulator 68000 : Step 1 instruction
*/
void EMU68_step(void)
{
  u32 line,opw,reg9;
#ifdef EMU68DEBUG
  opw = (u16)get_nextw();
#else
  u8 *mem = reg68.mem + (reg68.pc&MEM68MSK);
  reg68.pc += 2;
  opw  = (mem[0]<<8) | mem[1];
#endif
  line = opw&0xF000;
  opw -= line;
  reg9 = opw&0x0E00;
  opw -= reg9;
  line |= opw<<3;
  reg9 >>= 9;
  line >>= 6;
  opw  &=  7;
  (line_func[line])(reg9,opw);
}

static void poll_rte( u32 stack )
{
  do
  {
    EMU68_step();
  } while((unsigned)reg68.a[7] < stack);
}

/* Run until RTS level 0 occurs
 * and play Interruption for 1 pass.
 * ( Specific SC68 !!! )
*/
void EMU68_level_and_interrupt(u32 cycleperpass)
{
  u32 stack=reg68.a[7], pc=reg68.pc, cycle, fd;
  io68_t *io;
  int68_t *t;

  // $$$
//  BREAKPOINT68
  reg68.a[7] -= 4;
  EMU68_poke(reg68.a[7]+0, 0x12);
  EMU68_poke(reg68.a[7]+1, 0x34);
  EMU68_poke(reg68.a[7]+2, 0x56);
  EMU68_poke(reg68.a[7]+3, 0x78);

  /* Clear ORed memory access flags ... */
#ifdef EMU68DEBUG
  reg68.framechk = 0;
#endif

  /* Adjust internal IO cycle counter & find MFP */
  for(io=reg68.iohead; io!=NULL; io=io->next)
    (io->adjust_cycle)(reg68.cycle);
  reg68.cycle = 0;

  /* Do until RTS */
  do
  {
    EMU68_step();
  } while(stack > (unsigned)reg68.a[7]);
  cycle=0;

#ifdef _DEBUG
  if (stack != (unsigned)reg68.a[7]) {
    BREAKPOINT68
  }
  if (reg68.pc != 0x12345678) {
    BREAKPOINT68
  }
#endif


/*goto skip_int;*/

  /* Get interrupt IO (MFP) if any */
  if(interrupt_io!=NULL) {
    int ipl = (reg68.sr & (7<<SR_IPL_BIT)) >> SR_IPL_BIT;
    /* $$$ HACK : mfp interupt at level 5 */
    if ( 6 > ipl ) {
      /* Spool interrupt */
      while(fd=(interrupt_io->next_int)(cycle), fd!=IO68_NO_INT)
      {
        /* ... This must not happen ...  */
        if( (int)fd < 0 ) {
          fd = 0;
#ifdef _DEBUG
          BREAKPOINT68;
#endif

        }
        /* ... */

        /* TIME travel to next interruption */
        cycle += fd;

        if(cycle>=cycleperpass) break;

        /* Execute mfp interrupt emulation */
        t=(interrupt_io->interupt)(cycle);
        if(t!=NULL)
        {
/*          BREAKPOINT68*/
          reg68.cycle=cycle;
          reg68.a[7] = stack; /* $$$ Safety net */
          reg68.pc = 0x12345678;
          EXCEPTION(t->vector,t->level);
          poll_rte(stack);
#ifdef _DEBUG
          if (reg68.pc != 0x12345678) {
            BREAKPOINT68;
          }
          if (reg68.a[7] != (int)stack) {
            BREAKPOINT68;
          }
          if (!(reg68.sr & 0x2000)) {
            BREAKPOINT68;
          }
#endif

        } else {
//          BREAKPOINT68;
        }
      }
    }
  }

  /* Restore A7, PC, and new cycle counter */
  reg68.a[7] = stack;
  reg68.pc = pc;
  reg68.cycle = cycleperpass;
}

/* Run emulation for given number of cycle
 */
void EMU68_cycle(u32 cycleperpass)
{
  unsigned int cycle;
  io68_t *io;

  /* Clear ORed memory access flags ... */
#ifdef EMU68DEBUG
  reg68.framechk = 0;
#endif

  /* Adjust internal IO cycle counter */
  for(io=reg68.iohead; io!=NULL; io=io->next)
    (io->adjust_cycle)(reg68.cycle);

  /* Do until RTS */
  cycle = reg68.cycle = 0;
  do
  {

    /* If not stopped, execute an instruction */
    if (!reg68.status) {
      EMU68_step();
      cycle += 4;
    }
    reg68.cycle = cycle;


    if (interrupt_io) {
      int68_t *t;
      int ipl;
      /*
      fd = interrupt_io->next_int(cycle);
      if (fd != IO68_NO_INT) {
        int68_t *t;

        if (reg68.status) {
          if (fd > 0) {
            cycle += fd;
            if (cycle >= cycleperpass) {
              cycle = cycleperpass;
              break;
            }
          }
        }
        */

        ipl = (reg68.sr & (7<<SR_IPL_BIT)) >> SR_IPL_BIT;
        /* $$$ HACK : mfp interupt at level 5 */
        if ( 6 > ipl ) {
            // $$$
          t = (interrupt_io->interupt)(cycle);
          if (t) {
            reg68.status = 0;
            EXCEPTION(t->vector,t->level);
            /* $$$ cycle should be added here for exception handle */
            //BREAKPOINT68
          }

        }
      //}
    }



  } while(cycle < cycleperpass);

  // $$$ This is hack : should be "cycle" and handle later
  cycle = cycleperpass;
#if 0
  if (interrupt_io) {
    int68_t *t;
    int ipl;
    ipl = (reg68.sr & (7<<SR_IPL_BIT)) >> SR_IPL_BIT;
    /* $$$ HACK : mfp interupt at level 5 */
    if ( 5 > ipl ) {
      t = (interrupt_io->interupt)(cycle);
      if (t) {
        reg68.status = 0;
        EXCEPTION(t->vector,t->level);
        /* $$$ cycle should be added here for exception handle */
        //BREAKPOINT68
      }
    }
  }
#endif


//  reg68.cycle = cycle;
  reg68.cycle = cycle;

}


/* Run until PC = breakpc
*/
void EMU68_break(u32 breakpc)
{
  while((unsigned)reg68.pc!=breakpc)
  {
    EMU68_step();
  }
}

/* EMU68 fisrt time init
 * maxmem : 68000 memory requested in byte ( power of 2 only !!! )
 *
 * returns 0=NO ERROR
*/
int EMU68_init(u32 maxmem)
{
  int i;

  /* CLear 68000 registers and all */
  memset(&reg68,0,sizeof(reg68));
  interrupt_io = NULL;

  /* Check memory bound 128Kb <= MEM REQUESTED <= 16MB */
  if(maxmem<(1<<17))
    return EMU68error_add("EMU68_init : memory amount must be >= 128Kb");
  else if(maxmem>(1<<24))
    return EMU68error_add("EMU68_init : memory amount must be < 16mb");
  for(i=0; i<24 && maxmem!=((unsigned)1<<i); i++);
  if(maxmem!=((unsigned)1<<i))
    return EMU68error_add("EMU68_init : memory amount must be a power of 2");

  /* Must be done before next test */
  reg68.memsz = maxmem;
  reg68.memmsk = maxmem-1;

  /*  This test is neccessary becoz of possible optimizations in
   *  EMU68 compilation !!!
   */
  if(MEM68MSK != maxmem-1)
    return EMU68error_add( "EMU68_init : memory amount incompatible"
                           " with EMU68 version");

  /* Alloc 68000 memory buffer */
  if(reg68.mem=(u8 *)malloc(maxmem), reg68.mem==NULL)
    return EMU68error_add("EMU68_init : 68000 memory allocation failed");
  memset(reg68.mem,0,maxmem);

#ifdef EMU68DEBUG
  /* Alloc 68000 debug-memory buffer (acces flags) */
  if(reg68.chk=(u8 *)malloc(maxmem),reg68.chk==NULL)
    return EMU68error_add("EMU68_init : 68000 check-memory allocation failed");
  memset(reg68.chk,0,maxmem);
#endif

  EMU68memory_init();
  EMU68_reset();
  return 0;
}

/** Clean exit.
 */
void EMU68_kill(void)
{
#ifdef EMU68DEBUG
  /* Alloc 68000 debug-memory buffer (acces flags) */
  if(reg68.chk) {
    free (reg68.mem);
  }
#endif

  if(reg68.mem) {
    free (reg68.mem);
  }
  interrupt_io = NULL;
  memset(&reg68,0,sizeof(reg68));
}


/* Call & check linkage error !!!
*/
#ifdef EMU68DEBUG
void EMU68_debugcompiled(void)
{
}

#endif

/* Check if Good EMU68 mode : return sizeof(reg68_t)
*/
unsigned EMU68_sizeof_reg68(void)
{
  return sizeof(reg68_t);
}
