/**
 * @ingroup   winamp68_devel
 * @file      about.c
 * @author    Ben(jamin) Gerard <ben@sashipa.com>
 * @date      20/10/2001
 * @brief     sc68 about dialog box.
 * @version   $Id$
 */

/*
 *                       sc68 - winamp input plugin
 *         Copyright (C) 2001 Ben(jamin) Gerard <ben@sashipa.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "sc68glabout.h"

#include <windows.h>
#include <mmreg.h>
#include <msacm.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <commctrl.h>
#include <time.h>

// GL includes
#include <gl/gl.h>
#include <gl/glu.h>

#include "resource.h"
#include "in2.h"
#include "frontend.h" /* Winamp messages def */

#include "obj3d.h"
#include "version68.h"

#include "sc68app/SC68app.h"
#include "file68/SC68utils.h"
#include "file68/SC68os_general.h"
#include "io68/ymemul.h"

static In_Module * mod;               // the output module (declared in winamp68.c)
static u8 * paula;                    // Paula regiter data storage
static u8 (*YM_readreg_func)(u8,u32); // Read YM register function
static char * (*os_resource_name)(int, char *);

BOOL CALLBACK DialogProc(HWND,UINT,WPARAM,LPARAM);

static volatile HWND gHdlg = 0;

#define STD_FONT_LIST     0x100
#define BIG_FONT_LIST     0x200
#define URL_FONT_LIST     0x300
#define MAX_GLYPH         255

#define STD_FONT          0
#define BIG_FONT          1
#define URL_FONT          2

#define NB_GHOST    6
#define MAX_STARS   1024

#define HRULEZ(N)     ((((N)&0x3)>>0)-1)
#define VRULEZ(N)     ((((N)&0xC)>>2)-1)
#define SIZERULEZ(N)  (((N)>>8)&1)
#define COLORRULEZ(N) (((~(N))>>12)&7)
#define LIGHTRULEZ(N) (((N)>>16)&255)

// Animated Text
typedef struct {
  const char *text; // Text data
  unsigned int duration;   // Display duration (in mssec)
  unsigned int startTime;  // Start time stamp
} animtext_t;

typedef struct {
  const char *text;
  int   len;
  int   y;
  RECT  rect;
  int   rgb;
  int   size;
} textframe_t;

typedef struct {
  int width;
  int height;
  int xorig;
  int yorig;
  int xmove;
  int ymove;
} glyph_t;

typedef struct {
  HFONT hfont;  // Font handle
  int list;     // GL list id
  int ascent;
  int descent;
  glyph_t glyph[MAX_GLYPH];
} font_t;


// One star data
typedef struct {
  float x1,y1,z1,w1;
  float ghost[NB_GHOST][4];
} star_t;

// About instance data
typedef struct {
  // Windows TM
  RECT  r;
  HDC   hdc;
  HGLRC hglrc;
  HWND  winamp;
  int   width;
  int   height;

  // Font
  font_t font[3]; // 0:std, 1:big,  2:url

  // State
  int openGL;     // openGl correctly initialized
  unsigned int time, smoothTime;
  int inURL;      // Mouse in website url text bound
  //char *name;     // Window name

  int rundemo;
  int rundemoOk;

  SC68app_t *app;

  // Volume buffer
  int volume[256];
  int volIdx;

  // Anim text
  animtext_t *animtext, *curtext;
  unsigned int textWaitStamp;
  int nbText;
  textframe_t tf[128];
  int tfcnt;


  // Stars
  float ax,ay,az;       // Current angle
  float pax, pay, paz;  // Current angle step
  float gax, gay, gaz;  // Goal angle step
  float spd, gspd;      // Z move speed current & goal

  unsigned int starChgStamp;

  float zFar;
  int curGhost;
  int nbStars;
  star_t stars[MAX_STARS];

} about_t;


static void OutputSystemError(const char *fn, int line, const char  *errStr)
{
  char tmp[1024];
  int err = GetLastError();

  if (fn) {
    OutputDebugString(fn);
    sprintf(tmp,"(%d) : ",line);
    OutputDebugString(tmp);
  }

  tmp[0] = tmp[1023]=0;
  FormatMessage(
    FORMAT_MESSAGE_FROM_SYSTEM,      // source and processing options
    0,   // pointer to  message source
    err,  // requested message identifier
    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // language identifier for requested message
    tmp,    // pointer to message buffer
    sizeof(tmp)-1,        // maximum size of message buffer
    0  // pointer to array of message inserts
  );

  if (errStr) {
    OutputDebugString(errStr);
  }
  OutputDebugString(tmp);
  OutputDebugString("\n");
}

static void Clean(void)
{
  mod = 0;
  gHdlg = 0;
  paula = 0;
  os_resource_name = 0;
  YM_readreg_func = 0;
}

// Only exported function. We don't really need an header file for this !
SC68GLABOUT_API int CreateAboutDialog(int rundemo, sc68glabout_parm_t *p)
{
  if (!gHdlg) {
    if (p) {
      about_t * about = (about_t *)calloc(1, sizeof(about_t));
      if (about) {
        about->rundemo = rundemo;
        about->app = p->app;
        mod = p->in_mod;
        paula = p->paula;
        os_resource_name = p->rscname;
        YM_readreg_func = p->ymread;
        gHdlg = CreateDialogParam(mod->hDllInstance, (LPCTSTR)IDD_ABOUT, (HWND)p->hwnd,
                                 (DLGPROC)DialogProc,(LPARAM) about);
        if (!gHdlg) {
          free (about);
        }
      }
    }
  } else {
    ShowWindow(gHdlg,SW_SHOW);
  }
  if (!gHdlg) {
    Clean();
    return -1;
  }
  return 0;
}

static int MakeFontInfo(font_t *font, HDC hdc, HFONT hfont, int list)
{
  int c;
  TEXTMETRIC tm;
  ABC abc[MAX_GLYPH];
  int abcOk=1;

  memset(font,0, sizeof(*font));
  font->hfont = hfont;
  font->list = list;
  hfont = SelectObject(hdc, hfont);
  if (!hfont) {
    OutputSystemError(__FILE__,__LINE__,0);
    return -1;
  }
  wglUseFontBitmaps (hdc, 0, MAX_GLYPH-1, list);

  if(!GetCharABCWidths(hdc,0,127,abc)) {
    abcOk = 0;
    OutputSystemError(__FILE__,__LINE__,0);
  }

  if (!GetTextMetrics(hdc, &tm)) {
    OutputSystemError(__FILE__,__LINE__,0);
    font->ascent  = 7;
    font->descent = 3;
  } else {
    font->ascent = tm.tmAscent;
    font->descent = tm.tmDescent;
  }

  for (c=0; c<MAX_GLYPH; ++c) {
    static MAT2 mtx = { {1,1} , {1,0} , {1,0} , {1,1} };
    GLYPHMETRICS gm;
    if (GetGlyphOutline(hdc, c, GGO_BITMAP/*GGO_METRICS*/, &gm, 0, 0, &mtx) == GDI_ERROR) {
      OutputSystemError(__FILE__,__LINE__,0);
      return -1;
    }

    font->glyph[c].width  = gm.gmBlackBoxX;
    font->glyph[c].height = gm.gmBlackBoxY;
    font->glyph[c].xorig  = gm.gmptGlyphOrigin.x;
    font->glyph[c].yorig  = gm.gmptGlyphOrigin.y;
    font->glyph[c].xmove  = gm.gmCellIncX;
    font->glyph[c].ymove  = gm.gmCellIncY;

    // This looks better !
    if (abcOk) {
      font->glyph[c].xmove  = abc[c].abcA+abc[c].abcB+abc[c].abcC;
    }

  }
  if(hfont) SelectObject(hdc, hfont);
  return 0;
}

static int textColors[] = { 0x000000, 0xA0A0FF, 0xA0FFA0, 0xA0FFFF,
                            0xFFA0A0, 0xFFA0FF, 0xFFFFA0, 0xFFFFFF };

#ifdef _DEBUG

// $$$ FOR DEBUG
HBRUSH rBrush;
HBRUSH bBrush;
HBRUSH gBrush;
HBRUSH wBrush;
HPEN   wPen;

static void CreateDebugBrush(void)
{
  LOGBRUSH lb;

  lb.lbStyle = BS_SOLID;
  lb.lbHatch = 0;

  lb.lbColor = 0x00FF0000;
  bBrush = CreateBrushIndirect(&lb);

  lb.lbColor = 0x0000FF00;
  gBrush = CreateBrushIndirect(&lb);

  lb.lbColor = 0x000000FF;
  rBrush = CreateBrushIndirect(&lb);
  lb.lbColor = 0x00FFFFFF;
  wBrush = CreateBrushIndirect(&lb);

  wPen = CreatePen(PS_SOLID,0,0xFFFFFF);
}
#endif

typedef float matrix_t[4][4];

static void MtxIdentity(matrix_t m)
{
  int i,j;
  for (i=0; i<4; ++i) for (j=0; j<4; ++j) m[i][j] = (float)(i==j);
}

static void MtxMult(matrix_t m, const matrix_t m2)
{
  int i,k;
  for (i=0; i<4; ++i) {
    const float x = m[i][0];
    const float y = m[i][1];
    const float z = m[i][2];
    const float w = m[i][3];
    for (k=0; k<4; ++k) {
      m[i][k] = x * m2[0][k] + y * m2[1][k] + z * m2[2][k] + w * m2[3][k];
    }
  }
}

static void MtxVectMult(float *v, const float *u, const matrix_t m)
{
  const float x=u[0],y=u[1],z=u[2],w=u[3];
  v[0] = x * m[0][0] + y * m[1][0] + z * m[2][0] + w * m[3][0];
  v[1] = x * m[0][1] + y * m[1][1] + z * m[2][1] + w * m[3][1];
  v[2] = x * m[0][2] + y * m[1][2] + z * m[2][2] + w * m[3][2];
  v[3] = x * m[0][3] + y * m[1][3] + z * m[2][3] + w * m[3][3];
}


static void MtxTranspose(matrix_t m)
{
  int i, j;
  for (i=0; i<4; ++i) {
    for (j=i+1;j<4; ++j) {
      float tmp = m[i][j];
      m[i][j] = m[j][i];
      m[j][i] = tmp;
    }
  }
}

static void MtxRotateX(matrix_t m, const float a)
{
  const float ca = (float)cos(a);
  const float sa = (float)sin(a);
  matrix_t m2;
  MtxIdentity(m2);
  m2[1][1] = m2[2][2] = ca;
  m2[1][2] = sa;
  m2[2][1] = -sa;
  MtxMult(m,m2);
}

static void MtxRotateY(matrix_t m, const float a)
{
  const float ca = (float)cos(a);
  const float sa = (float)sin(a);
  matrix_t m2;
  MtxIdentity(m2);
  m2[0][0] = m2[2][2] = ca;
  m2[2][0] = sa;
  m2[0][2] = -sa;
  MtxMult(m,m2);
}

static void MtxRotateZ(matrix_t m, const float a)
{
  const float ca = (float)cos(a);
  const float sa = (float)sin(a);
  matrix_t m2;
  MtxIdentity(m2);
  m2[0][0] = m2[1][1] = ca;
  m2[0][1] = sa;
  m2[1][0] = -sa;
  MtxMult(m,m2);
}

static void CrossProduct(float *d, const float *v, const float * w)
{
  d[0] = v[1] * w[2] - v[2] * w[1];
  d[1] = v[2] * w[0] - v[0] * w[2];
  d[2] = v[0] * w[1] - v[1] * w[0];
}

static void FaceNormal(float *d, const vtx_t * v, const tri_t *t)
{
  const vtx_t *a=v+t->a,*b=v+t->b,*c=v+t->c;
  float n;
  vtx_t A,B;

  A.x = a->x - b->x;
  A.y = a->y - b->y;
  A.z = a->z - b->z;

  B.x = a->x - c->x;
  B.y = a->y - c->y;
  B.z = a->z - c->z;

  CrossProduct(d,&A.x,&B.x);
  n = 1.0f / (float)sqrt((double)(d[0]*d[0] + d[1]*d[1] + d[2]*d[2]));
  d[0] *= n;
  d[1] *= n;
  d[2] *= n;
}

static void ResizeAndCenter(obj_t *o, const float w)
{
  float m[2][3];
  float c[3];
  float s;
  int i,j;

  if (o->nbv != o->static_nbv || o->nbf != o->static_nbf) {
    // !!! Problem in genenerated object file
    BREAKPOINT68;
  }

  if (o->flags&1) {
    return;
  }
  o->flags |= 1;

  // Set W
  for (i=0; i<o->nbv; ++i) {
    o->vtx[i].w = 1.0f;
  }


  // Set first min-max
  for (j=0; j<3; ++j) {
    m[0][j] =
    m[1][j] = (&o->vtx[0].x)[j];
  }

  // Scans others
  for (i=1; i<o->nbv; ++i) {
    for (j=0; j<3; ++j) {
      const float v = (&o->vtx[i].x)[j];
      if (v < m[0][j]) {
        m[0][j] = v;
      } else if (v > m[1][j]) {
        m[1][j] = v;
      }
    }
  }

  // Calculates center
  for (j=0; j<3; ++j) {
    c[j] = 0.5f * (m[0][j] + m[1][j]);
  }

  // Calculates scale factor
  s = w / (m[1][0] - m[0][0]);
  for (j=1; j<3; ++j) {
    float s2 = w / (m[1][j] - m[0][j]);
    if (s2 < s) {
      s = s2;                                    }
  }

  // Transform all
  for (i=0; i<o->nbv; ++i) {
    float tmp;
    for (j=0; j<3; ++j) {
      (&o->vtx[i].x)[j] = ((&o->vtx[i].x)[j] - c[j]) * s;
    }
    tmp = o->vtx[i].z;
    o->vtx[i].z = o->vtx[i].y;
    o->vtx[i].y = tmp;
  }

  // Inverse face def
  for (i=0; i<o->nbf; ++i) {
    o->tri[i].a ^= o->tri[i].b;
    o->tri[i].b ^= o->tri[i].a;
    o->tri[i].a ^= o->tri[i].b;
  }

}

static animtext_t animText[] =
{
  /* OK */
  { "^-^^"
    "^1"  "Congratulations !" "^n^0"
    "You have successfully installed a" "^n"

    "^|"
    "^c0"
    "^1" "^c5" "sashipa & ben" "^0"
    "^c0" "production"

    "^v"
    "^c4"
    "^1"  "sc68"
    "^c1"
    "^0"  "version " WINAMP68_VERSION_STR
    "^c0"
    "Ultimate Atari-ST and Amiga soundchip emulator"

    , 8000 },

  /* OK */
  { "^-^^"
    "^1"  "About this intro" "^n^0"

    "^c4"
    "^n"  "Credits"
    "^c0" "Code and Art Works by Ben"
    "^c5" "<ben@sashipa.com>"
    "^c0" "Music by Frank Lautenbach (Dark Angel)"
    "^n"
    "^c0" "3D effects powered by OpenGL" "\256"
    "^n"
    "^n"
    "^c4" "Don't panic !" "^n^c0"
    "^<"  "   This introduction screen will not appear next time winamp will be "
          "launched. However, you can run it again from the [about] button "
          "in plugins preference menu."
    , 14000 },

  /* OK */
  { "^-^^"
    "^1" "Presentation" "^n^0"
    "^<^|"
    "^n"  "   sc68 is a player for Atari-ST & Amiga musics. It is the "
          "best emulated Atari-ST sound ever ! If you like old "
          "school chip-tunes, you will surely be fulfilled."

    "^n"  "   sc68 needs special music files (.sc68). A great collection of them "
          "could be found the the official website. Almost all Atari ST games and demos "
          "musics could be found there as well as Amiga chip tunes."

    "^>^v^c1"
          "Click below to open mini-browser"
    , 14000},

  /* OK */
  { "^-^^"
    "^1" "New features" "^n^0"
    "^<^v"
    "^c4"   "- This intro screen. Good looking, isn't it ?"
    "^c0"   "- Sid voices enhanced emulation. "
            "Tao and Scavenger tunes sound really great now !"
    "^c4"   "- Additionnal external replay"
    "^-^c0"  "Quartet (Rob Povey excellent soundtracker)"
    "\n"    "TCB tracker (Ancool/TCB soundtracker)"
    "\n"    "SidSoundDesigner (Scavenger replay updated)"
    "^c4"   " ... and more ..."
    , 12000},

  /* OK */
  { "^-^^"
    "^1" "Special Greetings" "^0"
    "^c2" "external contributions"
    "^|^-"
    "^c0"
    "Georges Kesseler (Gunstick / ULM)\n"
    "Mathieu Stempell (Dma-Sc / SectorOne)\n"
    "Nicholas Pomared (Mr.Stycks)\n"
    "Phil (Grazey / PHF)\n"
    "Patrick Bricout (Mr.Bee / OVR)\n"
    "Richard Davey\n"
    "Rob Perry"
    , 12000},

  /* OK */
  { "^-^^"
    "^1" "Greetings" "^0"
    "^c4" "Old Atari ST scene"

    "^v"
    "^c1" "^<" "WizzFrogs" "^-"
    "^c0" "Overlanders ~ St Connexion ~ Legacy ~ Naos" "^n"
    "^c2" "^<" "The Union" "^-"
    "^c0" "Tex ~ TCB ~ Delta force ~ The Replicants ~ Level 16 ~ ULM ~ The Respectable"
    "^c5" "^<" "Lonesome coders" "^-" "^n"
    "^c0" "Mcoder ~ Gudul"
    "^c6" "^<" "More great crews" "^-"
    "^c0" "The Lost boys ~ Next ~ Oxygene ~ Holocaust ~ Inner Circle ~ ..."
    ,10000 },

  /* OK */
  { "^-^^"
    "^1" "Greetings" "^0"

    "^c2" "Soundchip musicians"
    "^v"

    "^c0"
"505 ~ ACC ~ AJT ~ Aaron Fothergill ~ Adrian Dalecki ~ Agrajac ~ Alex Herbert ~ "
"Alister Brimble ~ Anders Nilson (An Cool) ~ Andi McGinty ~ Andrew Taylor ~ "
"Anthony Lees ~ Anti Christ ~ Ash Hogg ~ Atomic - Albedo ~ Barry Leitch ~ "
"Beast ~ Ben Daglish ~ BuZz ~ Celtic ~ Chris Huelsbeck ~ Chris Mad ~ "
"Chrispy Noodle / BBC ~ Christopher Guard ~ Crawdaddy ~ Cyril Payet (Doclands) ~ "
"Cytax ~ Damon Chaplin ~ Dave Kelly ~ Dave Lowe ~ Dave Moss (Spaz)"
    "^>^v"
    "^c4" "more ..."
    , 10000 },


  /* OK */
  { "^-^^"
    "^1" "Greetings" "^0"

    "^c6" "Soundchip musicians"
    "^v"

    "^c0"
"Dave Rogers ~ David Whittaker ~ Demon X ~ Deneb ~ Dextrous ~ Digi Tallis ~ "
"Doh ~ Donovan Prince ~ Dr B (Galtan 6) ~ Dr. Awesome ~ ENS ~ Emmanuel Larry (Lap) ~ "
"Epsilon & Epic (TMS) ~ Excellence In Art ~ Frank Lautenbach (Dark Angel) ~ "
"Frazer ~ Fred Gray ~ Gunnar Gaubatz (Big Alec) ~ Holger Gehrmann ~ Hydra ~ Iso ~ "
"J.C. Brooke ~ J.S. Gerard (Jess) ~ Janko Mrsic (Flogel) ~ Jason Page ~ "
"Jochen Hippel (Mad Max) ~ John M. Phillips (JMP) ~ Jonathan Dunn"
    "^>^v"
    "^c4" "more ..."
    , 10000 },

  /* OK */
  { "^-^^"
    "^1" "Greetings" "^0"

    "^c1" "Soundchip musicians"
    "^v"

    "^c0"
"Jurgen Piscol (Sound Machine) ~ Justin Scharvona ~ Karsten Obarski ~ Keith Tinman ~ "
"Keo ~ Kerni ~ LTK ~ Ladislav John Brozovski ~ Laurens Tummers (Lotus) ~ M. Lundberg ~ "
"MSG - RG ~ Maniacs Of Noise ~ Mark Cooksey ~ Mark Wilson ~ Martin Walker ~ "
"Mathieu Stempell (Dma-Sc) ~ Matt Furniss ~ Matthew Simmonds ~ McLaser ~ Mike Davies ~ "
"Mookie ~ Neil Crossley ~ Nic Alderton (Count Zero) ~ Nick Jones ~ Nigel Pritchard"
    "^>^v"
    "^c4"
    "Did I forget someone ?"
    , 10000 },

  /* OK */
  { "^-^^"
    "^1" "Greetings" "^0"

    "^c4" "Soundchip musicians"
    "^v"

    "^c0"
"No More ~ Paul Loraine ~ Paul Shields ~ Paul Tonge & Tony Williams ~ Pierre Loriaux ~ "
"Powa ~ Professional Sound Artists ~ Rhino ~ Richard Joseph ~ Richard Lopes (Griffon) ~ "
"Rob Hubbard ~ Rob Perry ~ Scavenger ~ Sean Connoly ~ Sean Conran ~ Sector One ~ "
"Shaft ~ Skyline ~ Sparky - Pompey Pirates ~ Stefan Jeworowski ~ "
"Stephen Mifsud (FantaZY) ~ Steve Barett ~ Steve Wheterill ~ Tao ~ The Innovator ~ "
"The Mind ~ Tim Follin ~ Titan ~ Tony Williams ~ Tristar ~ Tyan ~ Ultima ~ "
"Veggy (Dr B) ~ Wally Beben ~ Xag"
    "^>^v"
    "^c4"
    "Did I forget someone ?"

    , 10000 },

  /* OK */
  { "^-^^"
    "^1" "Personnal Greetings" "^0"

    "^c1" "sashipa fellows and friends"
    "^|^-"
    "^c0"
    "Alex ~ "
    "Captain Filoret ~ "
    "Cha ~ "
    "David Mistouflet (Dim / OVR) ~ "
    "Fab (Fury / Legacy) ~ "
    "Fgr ~ "
    "Fred (DTX sync. / Legacy) ~ "
    "Gillou ~ Grand Filou ~ "
    "Kral ~ "
    "Lea ~ "
    "Let ~ "
    "Lia ~ "
    "Lo ~ "
    "Melo ~ "
    "Nic ~ "
    "Patrice (Furyo / OVR) ~ "
    "Patrick Bricout (Mr Bee / OVR) ~ "
    "Pauline ~ "
    "Pib (DRD) ~ "
    "RV ~ "
    "Thomas ~ "
    "Vince ~ "
    "Vincent Penne (Ziggy Stardust / OVR)"
    "^c4"
    "... and all others I forgot ..."
    , 12000 },

  /* OK */
  {
    "^-^^^c0"
    "Did you notice these marvellous colors ?"
    "^v"

    "^c0" "White"
    "^c1" "Cyan"
    "^c2" "Magenta"
    "^c3" "Blue"
    "^c4" "Yellow"
    "^c5" "Green"
    "^c6" "Red"
    "^c7" "Black"
    "^c0" "Believe it or not but the color of line above is black !"
    ,6000 },

  /* OK */
  {
    "^<^^" "^1" "Atari logo"
    "^-^|"
    "^0"
    "\n" "Weird"
    "\n" "thing"
    "\n" "to see"
    "\n" "<- these 2 ->"
    "\n" "<- logos ->"
    "\n" "in the same"
    "\n" "screen, isn't it ?"
    "^>^v" "^1" "Amiga logo"
    , 8000 },


};

static float FloatRand(const float min, const float max)
{
  return ((float)(rand() & 0x7FFF) / (float)0x7FFF) * (max - min) + min;
}

static float SmoothFloat(const float a, const float factor)
{
  const float epsilon = 1E-4f;
  float r = a * factor;
  if (fabs(r) < epsilon) {
    r = 0.0f;
  }
  return r;
}

static float BlendFloat(const float a, const float b, const float r) {
  return (a*r) + (b*(1.0f-r));
}

static void ChangeStarMove(float * ax, float * ay, float * az, float *spd)
{
  const float max = 1.0f;
  *ax = FloatRand(-max, max);
  *ay = FloatRand(-max, max);
  *az = FloatRand(-max, max);
  *spd = FloatRand(2.0f, 8.0f);
}

static int SetWindowPixelFormat(HDC hDC, int gl, int doubleBuffer)
{
	PIXELFORMATDESCRIPTOR pixelDesc;
	int		m_GLPixelIndex;

  memset(&pixelDesc, 0, sizeof(pixelDesc));
	pixelDesc.nSize		= sizeof(PIXELFORMATDESCRIPTOR);
	pixelDesc.nVersion	= 1;
	pixelDesc.dwFlags	=	PFD_DRAW_TO_WINDOW |
                      (gl ? PFD_SUPPORT_OPENGL : 0)|
                      (doubleBuffer ? PFD_DOUBLEBUFFER : 0) |
							        PFD_STEREO_DONTCARE;
	pixelDesc.iPixelType = PFD_TYPE_RGBA;
	pixelDesc.cColorBits = 32;
	pixelDesc.cDepthBits = 16;
	pixelDesc.iLayerType = PFD_MAIN_PLANE;

	m_GLPixelIndex = ChoosePixelFormat( hDC, &pixelDesc);
  if (m_GLPixelIndex==0) {// Let's choose a default index.
		m_GLPixelIndex = 1;
	}
	if (!DescribePixelFormat(hDC,m_GLPixelIndex,sizeof(PIXELFORMATDESCRIPTOR),&pixelDesc)) {
    OutputSystemError( __FILE__ , __LINE__, "DescribePixelFormat:");
		return -1;
	}
	if (!SetPixelFormat(hDC,m_GLPixelIndex,&pixelDesc)) {
    OutputSystemError( __FILE__ , __LINE__, "DescribePixelFormat:");
		return -1;
	}

	return 0;
}

static int InitAboutData(HWND hdlg, about_t *about)
{
  const float l = 7.0f;
  const float s = 2.0f * l / (float)0x1000;
  int i;
  SC68app_t *app;
  int rundemo;

  if (!about) {
    return -1;
  }
  app = about->app;
  rundemo = about->rundemo;
  memset(about,0,sizeof(*about));
  about->app = app;
  about->rundemo = rundemo;
  about->winamp = FindWindow("Winamp v1.x",NULL);

#ifdef _DEBUG
  CreateDebugBrush();
#endif

  about->time = GetTickCount();
  about->smoothTime = 10; // 10 ms, is timer requested rate.

  about->animtext = animText;
  about->nbText = sizeof(animText) / sizeof(*animText);

  GetClientRect(hdlg, &about->r);

  // Init stars
  about->nbStars = MAX_STARS;
  about->zFar    = l;

  about->spd = about->gspd = 0;
  about->ax = about->ay = about->az =
  about->pax = about->pay = about->paz =
  about->gax = about->gay = about->gaz = 0.0f;
  about->starChgStamp = 0;

  for (i=0; i<about->nbStars; ++i) {
    int j;
    about->stars[i].x1 = s * (float)(rand()&0xFFF) - l;
    about->stars[i].y1 = s * (float)(rand()&0xFFF) - l;
    about->stars[i].z1 = s * (float)(rand()&0xFFF) - l;
    about->stars[i].w1 = 1.0f;

    for (j=0; j<NB_GHOST; ++j) {
      int k;
      for (k=0; k<4; ++k) {
        about->stars[i].ghost[j][k]= (&about->stars[i].x1)[k];
      }
    }
  }

  // 3D objects
  ResizeAndCenter(&atariLogo, 1.0f);
  ResizeAndCenter(&amigaLogo, 1.0f);

  return 0;
}

static int TextRect(HDC hdc, int x ,int y, const char *text, RECT * r)
{
  int len = strlen(text);
  r->left = x;
  r->top = y;
  DrawText(hdc,text,len, r, DT_CALCRECT|DT_NOPREFIX);
  return len;
}

static void CalcTextRect(RECT * r, const font_t *font, const char *text, int len)
{
  int x=0, y=0;
  memset(r, 0, sizeof(*r));

//  BREAKPOINT68;

  while (len-- > 0) {
    int sx,sy;
    int x2,y2;
    const glyph_t *g;
    g = &font->glyph[(*text++)&127];

    sx = x;
    sy = y;

    // Get hot point

    x2 = x + g->width - g->xorig;
    y2 = y + g->height - g->yorig;

    x = x2 - g->width;
    y = y2 - g->height;

    if (x  < r->left)   r->left   = x;
    if (x2 > r->right)  r->right  = x2;
    if (y  < r->top)    r->top    = y;
    if (y2 > r->bottom) r->bottom = y2;

    x = sx + g->xmove;
    y = sy + g->ymove;
  }
}

static int RGBfromRules(int rules)
{
  int light = LIGHTRULEZ(rules);
  int rgb   = textColors[COLORRULEZ(rules)];
  return (rgb & 0xFFFFFF) | (light<<24);
}


static const char *SkipSpace(const char * text, int len)
{
  int c;
  while (c=*text, (len>0 && isspace(c))) {
    ++text;
    --len;
  }
  return text;
}

static const char *ReadWord(const char * text, int len)
{
  int c;
  while (c=*text, (len>0 && isgraph(c))) {
    ++text;
    --len;
  }
  return text;
}

static void UnionTargetRect(RECT *r, const RECT *r2)
{
  if (r2->left   < r->left)   r->left   = r2->left;
  if (r2->top    < r->top )   r->top    = r2->top;
  if (r2->right  > r->right)  r->right  = r2->right;
  if (r2->bottom > r->bottom) r->bottom = r2->bottom;
}

static void RelativeRect(RECT *r, const RECT *r2)
{
  r->left   -= r2->left;
  r->right  -= r2->left;
  r->top    -= r2->top ;
  r->bottom -= r2->top ;
}

static void InvRelativeRect(RECT *r, const RECT *r2)
{
  r->left   += r2->left;
  r->right  += r2->left;
  r->top    += r2->top ;
  r->bottom += r2->top ;
}

static void AddWord(RECT *lineBox, RECT *wordBox)
{
  if (wordBox->top < lineBox->top) {
    lineBox->top = wordBox->top;
  }
  if (wordBox->bottom > lineBox->bottom) {
    lineBox->bottom = wordBox->bottom;
  }
  wordBox->left += lineBox->right;
  wordBox->right += lineBox->right;
  lineBox->right = wordBox->right;
}

static int ReadWords(textframe_t *frame, int maxFrame, const font_t *font, const char *text, int len, int *prules)
{
  int rules = *prules;
  const char * newText = SkipSpace(text,len);
  int curSize = SIZERULEZ(rules);
  int curColor = COLORRULEZ(rules);
  int wordInLine, wordLen;

  len -= newText - text;
  text = newText;
  newText = ReadWord(text, len);
  wordLen = newText - text;
  for(wordInLine=0; wordLen && wordInLine < maxFrame; ++wordInLine) {
    len -= wordLen;
    frame[wordInLine].text = text;
    frame[wordInLine].len  = wordLen;
    frame[wordInLine].rgb  = RGBfromRules(rules);
    frame[wordInLine].size = curSize;
    CalcTextRect(&frame[wordInLine].rect, font + curSize, text, wordLen);
    frame[wordInLine].y = frame[wordInLine].rect.bottom;


    // Read next word
    newText = SkipSpace(text = newText, len);
    len -= newText - text;
    text = newText;
    newText = ReadWord(text, len);
    wordLen = newText - text;
  }
  return wordInLine;
}

static int FormatTextLine(textframe_t *frame, int maxFrame, RECT *r, const font_t *font, const char *text, int len, int *prules)
{
  textframe_t *saveframe = frame;
  // Working frame horizontal dimension
  const int frameL = r->left;
  const int frameR = r->right;
  const int frameW = frameR - frameL;

  int curSize = SIZERULEZ(*prules);

  int  wordInLine;  // Number of word in current line
  int spaceW;
  int lineY, lineX, lineAscent, lineDescent, firstLineAscent ;

  // Clear result frame
  memset(r,0,sizeof(*r));

  wordInLine = ReadWords(frame, maxFrame, font, text, len, prules);

  lineX = 0;
  lineY = 0;
  lineAscent  = 0;
  lineDescent = 0;
  firstLineAscent = 0;
  spaceW = 0;
  for (; wordInLine; --wordInLine, ++frame) {
    int w, moveX, moveY;

    w = lineX + spaceW + frame->rect.right; // Width when adding word
    if (lineX == 0 || w <= frameW) {
      moveX = lineX + spaceW;
    } else {
      moveX = 0;
      if (firstLineAscent < 0) {
        firstLineAscent = lineAscent;
      }
      lineY = lineY + lineAscent + lineDescent;
      lineDescent = lineAscent = 0;
    }

    if (font[frame->size].ascent > lineAscent) {
      lineAscent = font[frame->size].ascent;
    }
    if (font[frame->size].descent> lineDescent) {
      lineDescent = font[frame->size].descent;
    }

    spaceW = (font[frame->size].glyph[32].xmove+2);// + font[frame->size].glyph[frame->text[frame->len-1]&255].xmove;
    moveY = lineY;
    frame->rect.left    += moveX;
    frame->rect.right   += moveX;
    frame->rect.top     += moveY;
    frame->rect.bottom  += moveY;
    lineX = frame->rect.right;
    UnionTargetRect(r, &frame->rect);
  }
  wordInLine = frame-saveframe;

  firstLineAscent = firstLineAscent - r->top;
  for (frame = saveframe; wordInLine; --wordInLine, ++frame) {
    frame->rect.top     += firstLineAscent;
    frame->rect.bottom  += firstLineAscent;
  }
  r->top    += firstLineAscent;
  r->bottom += firstLineAscent;

  return frame-saveframe;
}


static void AlignAxe(int *l, int *r, int l2, int r2, int alignRules)
{
  int l1 = *l, r1 = *r;
  int off;

  if (alignRules < 0) {
    // left
    off = l2-l1;
  } else if (!alignRules) {
    // center
    off = (l2-l1+r2-r1)>>1;
  } else {
    // right
    off = r2-r1;
  }
  *l = l1+off;
  *r = r1+off;
}

static void CenterRect(RECT *r, const RECT *r2)
{
  AlignAxe(&r->left, &r->right, r2->left, r2->right,0);
  AlignAxe(&r->top, &r->bottom, r2->top, r2->bottom,0);
}

static void AlignRect(RECT *r, const RECT *r2, int hRules, int vRules)
{
  AlignAxe(&r->left, &r->right,  r2->left, r2->right,  hRules);
  AlignAxe(&r->top,  &r->bottom, r2->top,  r2->bottom, vRules);
}

static const char *GetLine(const char *text)
{
  int c;
  for (c=*text; c && c != '^' && c != '\n'; c=*(++text));
  return text;
}

static int rFormatText(HDC hdc, font_t * font, const RECT * winRect, RECT * blkRect, const char *text, int rules, textframe_t *tf, int maxtf)
{
  RECT selfRect;
  int cnt = 0;
  int len;
  int newRules = rules;
  int c;
  int lineFrames;
  const int vsep = 0;
  const char *nxtText = 0;

  // Parse text command
  while (text) {
    c = *text;
    if (c == 0) {
      text = 0;
    } else if (c == '\n') {
      ++text;
    } else if (c == '^') {
      switch(*(++text)) {
#ifdef _DEBUG
        case 'b': BREAKPOINT68; break;
#endif
        case '<': newRules = (newRules&~0x3) | (0<<0); break; // Align left
        case '-': newRules = (newRules&~0x3) | (1<<0); break; // Horizontal align center
        case '>': newRules = (newRules&~0x3) | (2<<0); break; // Align right
        case '^': newRules = (newRules&~0xC) | (0<<2); break; // Align top
        case '|': newRules = (newRules&~0xC) | (1<<2); break; // Vertical align center
        case 'v': newRules = (newRules&~0xC) | (2<<2); break; // Align bottom
        case '0': case '1':
          // Change font size
          newRules = (newRules & ~0x0F00) | (((*text)-'0')<<8);
          break;
        case 'n': // New paragraph
          blkRect->bottom += 4;
          break;
        case 'c': // New Color
          newRules = (newRules & ~0xF000) | ((*(++text) - '0')<<12);
          break;

      }
      ++text;
    } else {
      break;
    }
  }

  if (!text) {
    return 0;
  }

  // Check for new vertical rules.
  // A vertical alignment change cause a new block to be created.
  // !!! Window  could be modify or not... This time it is not.
  if ( (rules & 0xC) != (newRules & 0xC) ) {
    //RECT newWinRect;
    RECT newBlkRect;
    memset(&newBlkRect,0,sizeof(newBlkRect));
    // Current block must be verticaly align on window.
    // At this point horizontal alignment is not important
    AlignRect(blkRect, winRect, -1, VRULEZ(rules));

    /*
    newWinRect.left   = winRect->left;
    newWinRect.right  = winRect->right;
    newWinRect.bottom = winRect->bottom;
    newWinRect.top    = blkRect->bottom;
    if (newWinRect.top  > newWinRect.bottom) {
      BREAKPOINT68
      newWinRect.top = newWinRect.bottom;
    }
    FrameRect(hdc, &newWinRect, bBrush);
    rFormatText(hdc, font, &newWinRect, &newBlkRect, text, newRules);
    */

    return rFormatText(hdc, font, winRect, &newBlkRect, text, newRules, tf, maxtf);
  }

  nxtText = GetLine(text);
  len = nxtText-text;

  selfRect.left = selfRect.top = selfRect.bottom = 0;
  selfRect.right = winRect->right - winRect->left;
  lineFrames = FormatTextLine(tf, maxtf, &selfRect, font, text, len, &newRules);

  // Append this line below current block
  blkRect->bottom += 3;
  selfRect.top += blkRect->bottom;
  blkRect->bottom = selfRect.bottom += blkRect->bottom;

  // Align to line to the left of block
  AlignAxe(&selfRect.left, &selfRect.right, blkRect->left, blkRect->right, -1);

  // Enlarge block rectangle if neccessary
  if (selfRect.right > blkRect->right) blkRect->right = selfRect.right;

  // Make self relative to block left/top corner
  RelativeRect(&selfRect, blkRect);

  // Continue format
  cnt = rFormatText(hdc, font, winRect, blkRect, nxtText, newRules, tf+lineFrames, maxtf-lineFrames);

  // Align block in window
  AlignRect(blkRect, winRect, HRULEZ(newRules), VRULEZ(newRules));

  // Make self inside block
  InvRelativeRect(&selfRect, blkRect);

  // Horizontal align in block
  AlignAxe(&selfRect.left, &selfRect.right, blkRect->left, blkRect->right, HRULEZ(newRules));
  //FrameRect(hdc, blkRect, rBrush);

  // Draw it

  cnt += lineFrames;
  maxtf -= lineFrames;
  for( ;lineFrames; --lineFrames, ++tf) {
    tf->rect.left += selfRect.left;
    tf->rect.right += selfRect.left;
    tf->rect.top += selfRect.top;
    tf->rect.bottom += selfRect.top;
  }

//  PlanarToSpace(int x, int y, int width, int height, float *px, float *py);
//  DrawGlRect(const RECT *r, int rgb, int width, int height)

  return cnt;
}

static int FormatText(HDC hdc, font_t * font, const RECT * winRect, const char *text, int rules, textframe_t *tf, int maxtf)
{
  int cnt;
  RECT localr;
  memset(&localr,0,sizeof(RECT));
  //FrameRect(hdc, winRect, rBrush);
  cnt = rFormatText(hdc, font, winRect, &localr, text, rules, tf, maxtf);
  //FrameRect(hdc, &localr, wBrush);
  return cnt;
}


static void PlanarToSpace(int x, int y, int width, int height, float *px, float *py)
{
  const float sx = (float) 2.0f / (float)width;
  const float sy = (float) 2.0f / (float)height;

  *px = sx * (float)(x - (width>>1));
  *py = sy * (float)((height-y) - (height>>1));
}

static void DrawGlRect(const RECT *r, int rgb, int width, int height)
{
  float x, y;
  float R,g,b;
  R = (float)((rgb>> 0)&255) / 255.0f;
  g = (float)((rgb>> 8)&255) / 255.0f;
  b = (float)((rgb>>16)&255) / 255.0f;

  glBegin(GL_LINE_LOOP);
  glColor3f(R,g,b);
  PlanarToSpace(r->left, r->top, width, height, &x, &y);
  glVertex2f(x,y);

  PlanarToSpace(r->right, r->top, width, height, &x, &y);
  glVertex2f(x,y);

  PlanarToSpace(r->right, r->bottom, width, height, &x, &y);
  glVertex2f(x,y);

  PlanarToSpace(r->left, r->bottom, width, height, &x, &y);
  glVertex2f(x,y);
  glEnd();
}


static DrawGlText(int list, const char *text, int len, const float x, const float y, int rgb)
{
  float R,g,b,a;
  R = (float)((rgb>> 0)&255) / 255.0f;
  g = (float)((rgb>> 8)&255) / 255.0f;
  b = (float)((rgb>>16)&255) / 255.0f;
  a = (float)((rgb>>24)&255) / 255.0f;

  // $$$ HACK : This changes the text color !
  glColor4f(R,g,b,a);
  glVertex2f(0,0);

  glRasterPos2f(x, y);
  glListBase(list);
  glCallLists(len, GL_UNSIGNED_BYTE, text);
}

static DrawURL(about_t *about, HWND h, HDC whdc)
{
  const char *t = "http://sashipa.ben.free.fr";
  RECT tr;
  int len = strlen(t);
  int rgb = 0xFFFF8080;
  int sbottom;
  float x,y;
  POINT pt;
  if (about->font[URL_FONT].hfont) {
    SelectObject(whdc, about->font[URL_FONT].hfont);
  }
  GetCursorPos(&pt);
  ScreenToClient(h,&pt);

  CalcTextRect(&tr, &about->font[URL_FONT], t, len);
  tr.right += 2;
  tr.bottom += 2;
  sbottom = tr.bottom;
  AlignRect(&tr, &about->r,1,1);
  if (pt.x >= tr.left && pt.x<tr.right &&
    pt.y >= tr.top  && pt.y<tr.bottom) {
    about->inURL = 1;
    rgb = 0xFFF0F0;
  }
  glDisable(GL_BLEND);
  PlanarToSpace(tr.left, tr.bottom-sbottom, about->width, about->height, &x, &y);
  DrawGlText(about->font[URL_FONT].list, t, len, x, y,rgb);
}

static void DrawObj(const obj_t * o, const float *pos, const float *angle)
{
  const float rad2deg = 57.2957795131f;
  const vtx_t *v = o->vtx;
  const tri_t *t = o->tri;
  int i;

	glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(0,0,-1.8f);
  if (pos) {
    glTranslatef(pos[0],pos[1],pos[2]);
  }
  if (angle) {
    glRotatef(rad2deg * angle[2], 0,0,1);
    glRotatef(rad2deg * angle[0], 1,0,0);
    glRotatef(rad2deg * angle[1], 0,1,0);
  }
  glBegin(GL_TRIANGLES);
  for (i=0; i < o->nbf; ++i) {
    float n[3];
    FaceNormal(n,v,t+i);
    glNormal3fv(n);
    glVertex3fv(&v[t[i].a].x);
    glVertex3fv(&v[t[i].b].x);
    glVertex3fv(&v[t[i].c].x);
  }
  glEnd();
}


static void DrawLogos(int curVolume, const float tsec)
{
  const float pi = 6.28318530718f;
  static float atariAngles[3] = { 0,0,0 };
  static float atariPos[4]    = { -1,0,0 };

  static float amigaAngles[3] = { 0,0,0 };
  static float amigaPos[4]    = { 1,0,0 };

  static float pa[3] = {0,0,0};
  static int ov[3] = {0,0,0};
  const int stpDw = 15;
  int c;
  const float sf = (float)pow(0.3, tsec);

  // Move
  for (c=0;c<3; ++c) {
    float smoothf = 0.25f;
    int v = (curVolume >> (c<<3)) & 255;
    if (v > ov[c]) {
      float delta = (float)(v - ov[c] + 257.0f) * (1.0f/512.0f);
      delta = (float)sqrt(delta);
      ov[c] = v;
      pa[c] += delta * 80.0f * tsec;  // impulse
    } else if (v<ov[c]) {
      smoothf = 0.1f;
      if (ov[c]>stpDw) {
        ov[c] -= stpDw;
      } else {
        ov[c] = 0;
      }
    }
    pa[c] = SmoothFloat(pa[c], (float)pow(smoothf,tsec)); // smooth
  }


  if (pa[0] == 0.0f) {
    atariAngles[1] = SmoothFloat(atariAngles[1], sf);
    amigaAngles[1] = SmoothFloat(amigaAngles[1], sf);
  } else {
    atariAngles[1] = (float)fmod(atariAngles[1] + pa[0] * tsec, pi);
    amigaAngles[1] = (float)fmod(amigaAngles[1] - pa[0] * tsec, pi);
  }

  if (pa[1] == 0.0f) {
    atariAngles[2] = SmoothFloat(atariAngles[2], sf);
  } else {
    atariAngles[2] = (float)fmod(atariAngles[2] + pa[1] * tsec, pi);
  }

  if (pa[2] == 0.0f) {
    amigaAngles[2] = SmoothFloat(amigaAngles[2], sf);
  } else {
    amigaAngles[2] = (float)fmod(amigaAngles[2] - pa[2] * tsec, pi);
  }

  DrawObj(&atariLogo, atariPos, atariAngles);
  DrawObj(&amigaLogo, amigaPos, amigaAngles);
}



static void SetLight(int id, const float *pos, const float *amb, const float *dif, const float *spe, const float *ate)
{
  id += GL_LIGHT0;
  glEnable(id);
  if (amb) glLightfv(id,GL_AMBIENT,amb);
  if (dif) glLightfv(id,GL_DIFFUSE,dif);
  if (spe) glLightfv(id,GL_SPECULAR,spe);
  if (pos) glLightfv(id,GL_POSITION,pos);
  if (ate) {
    glLightf(id,GL_CONSTANT_ATTENUATION,ate[0]);
    glLightf(id,GL_LINEAR_ATTENUATION,ate[1]);
    glLightf(id,GL_QUADRATIC_ATTENUATION,ate[2]);
  }
  glLightf(id,GL_SPOT_CUTOFF, 180);
  //DrawObj(&amigaLogo, pos, 0);
}

static int CALLBACK EnumFontFamExProc(
  ENUMLOGFONTEX *lpelfe,    // pointer to logical-font data
  NEWTEXTMETRICEX *lpntme,  // pointer to physical-font data
  int FontType,             // type of font
  LPARAM lParam             // application-defined data
)
{
  return 1;
}


static BOOL CALLBACK DialogProc(
  HWND hdlg,      // handle to dialog box
  UINT uMsg,      // message
  WPARAM wParam,  // first message parameter
  LPARAM lParam   // second message parameter
)
{
  about_t * about = (about_t *)GetWindowLong(hdlg, GWL_USERDATA);

  switch(uMsg) {

    case WM_ERASEBKGND: // Cause no erase
      return 1;

    case WM_TIMER:
      {
        InvalidateRect(hdlg, 0, 0);
      } return 0;

    case WM_PAINT:
      if (!about || !about->openGL) {
        return 0;
      } else {
        const unsigned int smoothFactor = 0x1;
        unsigned int time     = GetTickCount();
        unsigned int elapsed  = time - about->time;
        unsigned int smooth   = ((elapsed * smoothFactor ) + (about->smoothTime * (0x10-smoothFactor))) >> 4;
        const float  tsec     = (float) smooth * 0.001f;
        int width  = about->width = about->r.right - about->r.left;
        int height = about->height = about->r.bottom - about->r.top;
        const float aspect = (float)height / (float)width;
        int late = 0;
        int curVolume;

        // Load intro musique from sample resource directory if neccessary.
        if (about->rundemo && !about->rundemoOk && about->app && !about->app->ready) {
          COPYDATASTRUCT cds;
          char * rsc = os_resource_name(SC68rsc_sample, "Winamp - About Intro.sc68");

          cds.dwData = IPC_PLAYFILE;
          cds.lpData = (void *) rsc;
          cds.cbData = strlen((char *) cds.lpData)+1; // include space for null char

          if (!about->winamp) {
            about->winamp = FindWindow("Winamp v1.x",NULL);
          }
          if (about->winamp) {
            int ret;
            //ret = SendMessage(about->winamp, WM_COMMAND,WINAMP_BUTTON5,0);
            ret = SendMessage(about->winamp,WM_COPYDATA,(WPARAM)NULL,(LPARAM)&cds);
            ret = SendMessage(about->winamp,WM_WA_IPC,(WPARAM)cds.lpData,IPC_CHANGECURRENTFILE);
            ret = SendMessage(about->winamp,WM_WA_IPC,0,IPC_STARTPLAY);
            about->rundemoOk = 1;
          }
        }

        // Get prebuffering time
        if (mod && mod->outMod && mod->outMod->IsPlaying()){
          late = mod->outMod->GetWrittenTime() - mod->outMod->GetOutputTime();
          if (late < 0) late = 0;
        }

        // Read hardware volumes
        if (about->app && about->app->ready && about->app->cur_mus) {
          int v = 0, c;
          if (about->app->cur_mus->flags.ym) {
            for (c=0; c<3; ++c) {
              int v2 = YM_readreg_func(8+c,0)&0x1f;
              if (v2&0x10) v2 = 0xf;
              v |= v2 << (c<<3);
            }
            v |= v<<4;
          } else if (about->app->cur_mus->flags.amiga) {
            for (c=0; c<4; ++c) {
              int v2 = paula[0xA9 + (c<<4)] & 0x7f;
              if (v2>=0x40) v2 = 0x3f;
              v |= v2 << ((c<<3)+2);
            }
            v |= (v>>6)&0x03030303;
          }
          about->volIdx = (about->volIdx+1) & 255;
          about->volume[about->volIdx] = v;
          curVolume = about->volume[(about->volIdx - (late / 20)) & 255];
        } else {
          curVolume = 0;
        }

        // Suppose not in URL box.
        about->inURL = 0;
        about->time           = time;
        about->smoothTime     = smooth;

        // Set GL render state & clear buffers
        if (about->openGL) {
          glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
          glEnable(GL_DEPTH_TEST);
          glEnable(GL_CULL_FACE);
          glDisable(GL_BLEND);

          glCullFace(GL_BACK);

          glDrawBuffer(GL_BACK);
	        glClearColor(0,0,0,0);
  	      glClearDepth(1);
          glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

          glMatrixMode(GL_TEXTURE);
	        glLoadIdentity();

          glMatrixMode(GL_PROJECTION);
          glLoadIdentity();
          glFrustum(-1, 1, -aspect, aspect, 0.7f, 20);

	        glViewport(0, 0, width, height); // (0,0) is bottom left corner.
        }

        // Process lights
        if (about->openGL) {
          const float ambiant[] = { 0.25f, 0.25f, 0.25f, 1.0f};
          const float red[] = { 1.f, 0.0f, 0.0f, 1.0f};
          const float green[] = { 0.f, 1.0f, 0.0f, 1.0f};
          const float blue[] = { 0.f, 0.0f, 1.0f, 1.0f};

          const float diffuse[] = { 1.f, 0.0f, 0.0f, 1.0f};
          const float specular[] = { 0.0f, 0.0f, 1.0f, 1.0f};
          const float white[] = { 1.0f, 1.0f, 1.0f, 1.0f};
          const float gray50[] = { 0.5f, 0.5f, 0.5f, 1.0f};
          const float atenuation[] = { 0.0f, 1.0f, 0.0f};

          static float lightAngle;
          vtx_t lightV = { 2.0f, 0.0f, 0.0f, 1.0f};
          matrix_t m;

          MtxIdentity(m);
          MtxRotateZ(m, lightAngle );
          MtxRotateX(m, lightAngle * 0.87f);
          MtxRotateY(m, lightAngle * -0.733f);
          MtxVectMult(&lightV.x,&lightV.x, m);
          lightAngle += tsec * 0.2f;

          glEnable(GL_LIGHTING);
          glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, 1.0f); // Front Specular

          glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT  , ambiant);
          glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE  , gray50);
          glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR , gray50);
          glMaterialf (GL_FRONT_AND_BACK, GL_SHININESS , 64.0f);


          SetLight(0, &lightV.x, ambiant, green, white, 0/*atenuation*/);
          lightV.x = lightV.x;
          lightV.y = lightV.y;
          lightV.z = -lightV.z;
          SetLight(1, &lightV.x, ambiant, red, white, 0/*atenuation*/);

          glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, 1.0f); // Front Specular

        }

        if (about->openGL) {
          glDepthFunc(GL_LESS);

          // Draw atari & amiga logos
          DrawLogos(curVolume, tsec);

          // Draw stars
          {
            int i,ghost;
            star_t *s = about->stars;
            const float zStep = tsec * about->spd;
            const float zfar  = about->zFar;
            const float rad2deg = 57.30f * 1.2f;
            matrix_t mtx;

            if (time > about->starChgStamp) {
              ChangeStarMove(&about->gax, &about->gay, &about->gaz, &about->gspd);
              about->starChgStamp = time + 5000;
            }

            // Anim stars
            MtxIdentity(mtx);
            MtxRotateZ(mtx, about->az);
            MtxRotateX(mtx, about->ax);
            MtxRotateY(mtx, about->ay);
            about->ax += about->pax * tsec;
            about->ay += about->pay * tsec;
            about->az += about->paz * tsec;
            about->pax = BlendFloat(about->pax, about->gax, 0.99f);
            about->pay = BlendFloat(about->pay, about->gay, 0.99f);
            about->paz = BlendFloat(about->paz, about->gaz, 0.99f);
            about->spd = BlendFloat(about->spd, about->gspd, 0.99f);
            ghost = (about->curGhost+1) % NB_GHOST;

            about->curGhost = ghost;
            for (i = 0, s = about->stars; i<about->nbStars; ++i, ++s) {
              s->z1 += zStep;
              if (s->z1 <= zfar) {
                MtxVectMult(s->ghost[ghost], &s->x1, mtx);
              } else {
                int j;
                s->z1 -= 2*zfar;
                MtxVectMult(s->ghost[ghost], &s->x1, mtx);
                for (j=0; j<NB_GHOST; ++j) {
                  if (ghost != j) {
                    memcpy(s->ghost[j], s->ghost[ghost], sizeof(float)*4);
                  }
                }
              }
            }

            glDisable(GL_LIGHTING);
  	        glMatrixMode(GL_MODELVIEW);
            glLoadIdentity();
            glGetFloatv(GL_PROJECTION_MATRIX, mtx[0]);


            for (i=0,s = about->stars; i<about->nbStars; ++i, ++s) {
              float light = 0.0f;
              int j,k;
              glBegin(GL_LINE_STRIP);
              for (j=0, k=ghost; j < NB_GHOST; ++j) {
                if (k==ghost) {
                  float v[4];
                  MtxVectMult(v, s->ghost[k], mtx);
                  if (v[2] >= 0) {
                    light = v[2] / v[3];
                    light *= light;
                    light = 1.0f - 0.60f * light;
                  }
                } else {
                  light *= (1.0f - 1.0f / (float)NB_GHOST);
                }
                glColor3f(light,light,light);
                glVertex3fv(s->ghost[k]);

                if (--k < 0) k = NB_GHOST-1;
              }
              glColor3f(1,0,0);
              glEnd();
            }
          }
        }

        // Draw old-style equa :
        // Not very reactiv (becoz of winamp output buffer!)
        if (1) {
          static int ov[3] = {0,0,0};
          const float w = 0.03f, sep = w / 2.2f, hu = w/60;
          int c, v;
          int stepDw = elapsed;
          float x1,y1,x2,y2;
          const float z = 0.0f;
          const float colors[4][4] = {
            {1.0f,0.0f,0.0f,1.0f},  // Bottom Left
            {0.3f,1.0f,0.0f,1.0f},  // Bottom Right
            {0.8f,1.0f,0.0f,0.7f},  // Top Right
            {1.0f,1.0f,0.0f,0.7f},  // Top lLft
          };


          glDisable(GL_LIGHTING);
  	      glMatrixMode(GL_PROJECTION);
          glLoadIdentity();
  	      glMatrixMode(GL_MODELVIEW);
          glLoadIdentity();
          glDisable(GL_DEPTH_TEST);
          glEnable(GL_BLEND);
          glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);


          glBegin(GL_QUADS);
          x2 = -1.0f;
          for (c=0; c<4; ++c) {
            v = (curVolume >> (c<<3)) & 255;
            x1 = x2 + sep;
            x2 = x1 + w;
            y2 = -1.0f + 2.0f*sep;
            if ( v > ov[c]) {
              ov[c] = v;
            } else {
              ov[c]-=stepDw;
              if (ov[c]<v) {
                ov[c] = v;
              }
            }
            y1 = y2 + (float)(ov[c]) * hu;

            glColor4fv(colors[0]);
            glVertex3f(x1,y2,z);
            glColor4fv(colors[1]);
            glVertex3f(x2,y2,z);
            glColor4fv(colors[2]);
            glVertex3f(x2,y1,z);
            glColor4fv(colors[3]);
            glVertex3f(x1,y1,z);
          }
          glEnd();
        }

        {
          unsigned int etime;
          int rgb = 0xFF;
          animtext_t *curText;
          RECT win;

          // Waiting time
          if (about->textWaitStamp && time < about->textWaitStamp) {
            // Nothing to do !!!
            about->tfcnt = 0;
          } else {
            curText = about->curtext;
            if (curText == 0) {
              about->textWaitStamp = time + 3000;
              curText = about->animtext;
              curText->startTime = about->textWaitStamp;
              etime = 0;
            } else {
              etime = time - curText->startTime;
            }
            if (etime >= curText->duration) {
              curText++;
              about->textWaitStamp = time + 4000;
              if (curText >= about->animtext+about->nbText) {
                curText = about->animtext;
              }
              curText->startTime = about->textWaitStamp ;
              etime = 0;
            }
            about->curtext = curText;

            // Fade Fx.
            if (etime < 500) {
              rgb = etime * 255 / 500;
            } else if (curText->duration - etime < 500) {
              rgb = (curText->duration - etime) * 255 / 500;
            }

            win = about->r;
            win.left    += width>>3;
            win.right   -= width>>3;
            win.top     += height>>4;
            win.bottom  -= height>>3;

            about->tfcnt = FormatText(about->hdc, about->font, &win, curText->text,
                                      1 | (1<<2) | (rgb<<16),
                                      about->tf, sizeof(about->tf)/sizeof(*about->tf));
            /*{
              float x,y;
              PlanarToSpace(win.left,win.bottom, width, height, &x, &y);
              DrawGlRect(&win, 0xFF0000, width, height);
            }*/

          }
        }

        // Display formatted text
        if (1) {
          int i;

          glDisable(GL_LIGHTING);
  	      glMatrixMode(GL_PROJECTION);
          glLoadIdentity();
  	      glMatrixMode(GL_MODELVIEW);
          glLoadIdentity();
          glLoadIdentity();
          glEnable(GL_BLEND);
          glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

          for (i = 0; i < about->tfcnt; ++i) {
            float x,y;
            about->tf[i].rect.bottom -= about->tf[i].y;
            PlanarToSpace(about->tf[i].rect.left,about->tf[i].rect.bottom, width, height, &x, &y);
            //DrawGlRect(&about->tf[i].rect, 0xFFFF00, width, height);
            DrawGlText(about->font[about->tf[i].size].list,
                       about->tf[i].text, about->tf[i].len,
                       x, y, about->tf[i].rgb);

          }

        }

        if (1) {
          DrawURL(about, hdlg, about->hdc);
        }

        if (about->openGL) {
          glFlush();
          glFinish();
          SwapBuffers(about->hdc);
        }

        // Anim and Format text

        return 0;
      }

    case WM_LBUTTONUP:
      {
        if (about) {
          if (about->inURL) {
            const char *url = "http://sashipa.ben.free.fr/sc68";
            if (!about->winamp) {
              about->winamp = FindWindow("Winamp v1.x",NULL);
            }
            if (about->winamp) {
              if (SendMessage(about->winamp,WM_WA_IPC,0,IPC_INETAVAILABLE) == 1) {
                SendMessage(about->winamp,WM_WA_IPC,0,IPC_MBOPENREAL);
                SendMessage(about->winamp,WM_WA_IPC,(WPARAM)url, IPC_MBOPENREAL);
              }
            }
          }
        }
      } return 0;

    case WM_KEYUP:
      {
        int nVirtKey = (int) wParam;    // virtual-key code
        int lKeyData = lParam;          // key data
        if (nVirtKey == 27) {
          DestroyWindow(hdlg);
        }
      } return 0;

		case WM_INITDIALOG:
      {
        about = (about_t *)lParam;
        InitAboutData(hdlg, about);

        SetWindowLong(hdlg, GWL_USERDATA, (LONG)about);
        ShowWindow(hdlg, SW_SHOW);
        UpdateWindow(hdlg);
        SetTimer(hdlg,1,15,0);
      } return 1;

    case WM_SHOWWINDOW:
      {
        if (about) {
          about->openGL = 0;
	        about->hdc = GetDC(hdlg);
	        if (about->hdc  && !SetWindowPixelFormat(about->hdc,1,1))	{
	          about->hglrc = wglCreateContext(about->hdc);
	          if (wglMakeCurrent(about->hdc, about->hglrc)) {
		          about->openGL = 1;
	          }

            // Create font handles
            {
              LOGFONT lg;
              HFONT hf;
              memset(&lg,0,sizeof(lg));
              lg.lfCharSet = DEFAULT_CHARSET;
              strcpy(lg.lfFaceName,"Arial Black");
              lg.lfPitchAndFamily = 0;
              lg.lfWeight = FW_DONTCARE;
              lg.lfWidth = 0;
              lg.lfHeight = 16;
              lg.lfItalic = 0;
              lg.lfUnderline = 0;
              lg.lfStrikeOut = 0;
              lg.lfOutPrecision = OUT_DEVICE_PRECIS;
              lg.lfClipPrecision = CLIP_DEFAULT_PRECIS;
              lg.lfQuality = PROOF_QUALITY;
              lg.lfPitchAndFamily = FF_DONTCARE;
              lg.lfEscapement = 0;
              lg.lfOrientation = 0;

              if (!about->font[URL_FONT].hfont) {
                strcpy(lg.lfFaceName,"Arial");
                lg.lfUnderline = 1;
                lg.lfHeight = 14;
                hf = CreateFontIndirect(&lg);
                MakeFontInfo(&about->font[URL_FONT], about->hdc, hf, URL_FONT_LIST);
              }

              if (!about->font[BIG_FONT].hfont) {
                strcpy(lg.lfFaceName,"Arial");
                lg.lfUnderline = 0;
                lg.lfHeight = 24;
                hf = CreateFontIndirect(&lg);
                MakeFontInfo(&about->font[BIG_FONT], about->hdc, hf, BIG_FONT_LIST);
              }

              if (!about->font[STD_FONT].hfont) {
                strcpy(lg.lfFaceName,"Arial");
                lg.lfUnderline = 0;
                lg.lfHeight = 14;
                hf = CreateFontIndirect(&lg);
                MakeFontInfo(&about->font[STD_FONT], about->hdc, hf, STD_FONT_LIST);
              }

            }

	        }
        }



      } return 0;

    case WM_COMMAND:
      {
        int wNotifyCode = HIWORD(wParam); // notification code
        int wID = LOWORD(wParam);         // item, control, or accelerator identifier
        HWND hwndCtl = (HWND) lParam;     // handle of control

        switch(wID) {
          case IDOK:
            DestroyWindow (hdlg);
            break;
        }
      } return 0;


		case WM_DESTROY:
      if (about) {
        int i;
        for (i=0; i<sizeof(about->font)/sizeof(*about->font); ++i) {
          if (about->font[i].hfont) {
            DeleteObject(about->font[i].hfont);
            about->font[i].hfont = 0;
          }
        }

        if (about->hglrc) {
          wglDeleteContext(about->hglrc);
          about->hglrc = 0;
          about->openGL = 0;
        }

        free(about);
      }
      SetWindowLong(hdlg, GWL_USERDATA, 0); // Safe net
      gHdlg = 0;
#ifdef _DEBUG
      ///$$$ MEGA HACK
      //exit(0);
#endif
      return 0;

		case WM_CLOSE:
			DestroyWindow (hdlg);
			return 0;

  }
  return 0;
}
