/**
 * @ingroup   xmms_devel
 * @file      xmms68.c
 * @author    Ben(jamin) Gerard <ben@sashipa.com>
 * @date      2000
 * @brief     sc68 input plugin for xmms
 * @version   $Id$
 */
 
/*
 *                        sc68 - xmms input plugin
 *         Copyright (C) 2001 Ben(jamin) Gerard <ben@sashipa.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/* Standard includes */
#include <glib.h>
#include <pthread.h>
#include <string.h>
/*#include <math.h>*/
#include <stdio.h>

/* xmms includes */
#include "plugin.h"
#include "util.h"

/* generate config include */
#include "config.h"

/* SC68 includes */
#include "sc68app/SC68.h"
#include "file68/SC68error.h"
#include "file68/SC68utils.h"
#include "file68/SC68file.h"
#ifdef WITH_BROWSE68
# include "browse68/SC68browse.h"
#endif
#include "file68/SC68os_general.h"
#include "sc68app/SC68os_interface.h"
#include "sc68app/SC68play.h"
#include "sc68app/SC68delay.h"
#include "sc68app/SC68mixer.h"

/*
* 68000 emulator includes
*/
#include "emu68/emu68.h"
#include "emu68/ioplug68.h"
#include "io68/io68.h"

/* 512 Kb 68K memory buffer */
#define MEMORY (1<<19)

static InputPlugin xmms68_ip;
static gboolean going;
static gboolean audio_error;
static pthread_t play_thread;
static SC68app_t app;

/* $$$
 */
#define _(A) (A)

/** Display to output debug statcked error messages.
 */
static void spool_error_message(void)
{
  char *s;

  if (s = SC68error_get(), s != NULL) {
    SC68os_pdebug("Stacked Error Message:\n");
    do {
      SC68os_pdebug("%s\n", s);
    } while (s = SC68error_get(), s != NULL);
  } else
    SC68os_pdebug("No stacked error message\n");
}

/** Initialize 68000 emulator.
 */
static int init_68K(reg68_t * r, u32 sz)
{
  if (EMU68_init(sz) < 0) {
    return SC68error_add(EMU68error_get());
  }
  *r = reg68;
  r->sr = 0x2000;
  r->a[7] = sz - 4;
  if (YM_init() < 0)
    return SC68error_add("Yamaha 2149 emulator init failed");
  if (MW_init() < 0)
    return SC68error_add("Microwire emulator init failed");
  if (MFP_init() < 0)
    return SC68error_add("MFP emulator init failed");
  if (PL_init() < 0)
    return SC68error_add("Paula emulator init failed");
  return 0;
}

/** Set sc68 application to default values. */
static void set_sc68app_default(SC68app_t * app)
{
  memset(app, 0, sizeof(*app));
  app->cur_track = -1;
  SC68config_default(&app->config);
}

/** Display a simple about dialog. */
static void xmms68_about(void)
{
static GtkWidget *box;
static char * text = 
      "sc68 " VERSION " XMMS plugin\n"
      "\n"
#ifdef _DEBUG
      " !!! DEBUG VERSION !!! \n"
      "\n"
#endif       
      "Atari STF/E and Amiga sound emulator\n"
      "\n"
      "(c)1998-2001 sashipa and BeN(jamin) Gerard\n"
      "\n"
      "email : ben@sashipa.com\n"
      "\n";

  box = xmms_show_message(
    _("About SC68"),
    _(text),
  _("Ok"),
  FALSE, NULL, NULL);
  gtk_signal_connect(GTK_OBJECT(box), "destroy", gtk_widget_destroyed, &box);
}

/** Set XMMS info. */
static void xmms68_set_info(disk68_t *d, int track)
{
  music68_t *m;
  gchar *name = 0;
  unsigned int time;
  char alltime[32];

  if (track < 0) {
    track = -2;
    m = d->mus + d->default_six;
    time = d->time;
    if (d->nb_six > 1) {
      name = g_strdup_printf("%s[%02u] / %s", d->name, d->nb_six, m->author);
    } else {
      name = g_strdup_printf("%s / %s", d->name, m->author);
    }
  } else {
    m = d->mus + track;
    time = m->time;
    if (d->nb_six > 1) {
      name = g_strdup_printf("%02u %s / %s", track + 1, m->name, m->author);
    } else {
      name = g_strdup_printf("%s / %s", m->name, m->author);
    }
  }
  /* Set time to disk time ... for xmms display */
  xmms68_ip.set_info(name, d->time * 1000, 44100, 44100, 2);
  g_free(name);

  name = SC68utils_make_track_time_info(d->nb_six, d->time);
  strncpy(alltime, name, sizeof(alltime));
  name =
      g_strdup_printf("disk:%s [%s]  track:%s [%s]  "
                      "author:%s  composer:%s  "
                      "hardware:%s  "
                      "replay routine:%s"
                      " :-) ",
                      d->name, alltime, m->name,
                      SC68utils_make_track_time_info(track + 1, m->time),
                      m->author,
                      m->composer,
                      (m->flags.
                       ym ? (m->flags.
                             ste ? "Yamaha-2149 & MicroWire (STE)" :
                             "Yamaha-2149") : (m->flags.
                                               amiga ? "Paula (amiga)" :
                                               "unknown")),
                     m->replay ? m->replay : "internal"
                                               );
  xmms68_ip.set_info_text(name);
  g_free(name);
}


/** Test file format XMMS callback. */
static int xmms68_is_our_file(char *filename)
{
  return SC68app_verify_diskfile(filename) < 0 ? FALSE : TRUE;
}

#if 0
/* Things to do at end of track */
static int end_of_disk_reached(SC68app_t * app, int nxt_t)
{
  int err = 0;

  app->eod = 1;
  switch (app->eod_action) {
  case 0:                      /* No action */
    break;
  case 1:                      /* Stop */
    err = SC68stop(app);
    break;
  case 2:                      /* Next disk : Not implemented */
  case 3:
    if (nxt_t < 0)              /* No track to play !!! */
      err = SC68stop(app);
    else {
      err = SC68track(app, nxt_t);
      if (!err)
        err = SC68play(app);
    }
    break;
  default:
    err = SC68error_add("Unknown end of disk action (%d)", app->eod_action);
    break;
  }
  return err;
}
#endif

/** XMMS stop callback.
 */
static void xmms68_stop(void)
{
  if (going) {
    going = FALSE;
    pthread_join(play_thread, NULL);
    if (xmms68_ip.output) {
      xmms68_ip.output->close_audio();
    }
  }
}

static void *play_loop(void *arg)
{
  disk68_t *d = (disk68_t *)arg;

  if (!going || !app.ready || !d) {
    spool_error_message();
    going = FALSE;
    SC68app_desactiv(&app);
  }

  while (going) {
    unsigned int sign = SC68MIXER_CHANGE_SIGN;
    unsigned int cycle_this_pass = app.mix.cycleperpass;
    
    if( ! app.mix.buflen ) {
      EMU68_level_and_interrupt(cycle_this_pass);

      /* YM ??? */
      if (app.cur_mus->flags.ym)
        app.mix.buflen = YM_mix(cycle_this_pass);
      else
        app.mix.buflen = app.mix.stdbuflen;

      app.mix.buf = YM_get_buffer();

      /* MW ??? */
      if (app.cur_mus->flags.ste) {
        MW_mix(app.mix.buf, app.reg68.mem, app.mix.buflen);
        if (sign) {
          SC68mixer_stereo_16_LR(app.mix.buf,app.mix.buf, app.mix.buflen, sign);
        }
      }
      /* Paula ??? */
      else if (app.cur_mus->flags.amiga) {
        PL_mix(app.mix.buf, app.reg68.mem, app.reg68.mem + app.reg68.memsz,
               app.mix.buflen);

        SC68mixer_stereo_16_LR(app.mix.buf,app.mix.buf, app.mix.buflen,
                               SC68MIXER_CHANGE_SIGN);

        SC68mixer_blend_LR(app.mix.buf, app.mix.buf, app.mix.buflen,
                           app.mix.amiga_blend , sign);
      }
      /* Do stereo my self */
      else {
        SC68mixer_dup_L_to_R(app.mix.buf, app.mix.buf, app.mix.buflen, sign);
      }

      if (app.delay.onoff) {
        /* Apply delay ... */
        int delay;
        const int shift = 0;

        if (app.delay.time < 0)
          delay = (app.mix.buflen << shift) / -app.delay.time;
        else
          delay = (app.mix.buflen * app.delay.time) >> shift;
        if (!app.delay.lr)
          /* Delay ring */
          SC68delay(app.mix.buf, app.mix.buflen, delay, app.delay.strength, sign);
        else
          /* L/R spacializer delay */
          SC68delay_LR(app.mix.buf, app.mix.buflen, delay);
      }


      /* Advance time */
      SC68app_advance_time(&app);

      /* Reach end of track */
      if (app.time.elapsed > app.time.track) {
        if(app.cur_track+1 >= app.cur_disk->nb_six) {
          xmms68_stop();
        } else {
          SC68track(&app, app.cur_track+1);
          xmms68_set_info(app.cur_disk, app.cur_track);
        }
      }
    }

    if (going) {
      int tosend = app.mix.buflen << 2;

      xmms68_ip.add_vis_pcm(xmms68_ip.output->written_time(),
                            FMT_S16_LE, 2, tosend, app.mix.buf);
      while (xmms68_ip.output->buffer_free() < tosend && going) {
        xmms_usleep(30000);
      }
      xmms68_ip.output->write_audio(app.mix.buf, tosend);
      app.mix.buflen -= tosend >> 2; 
    }
  }
  xmms68_set_info(app.cur_disk, -1);
  SC68eject(&app);

  /* Make sure the output plugin stops prebuffering */
  xmms68_ip.output->buffer_free();
  xmms68_ip.output->buffer_free();

  going = FALSE;
  pthread_exit(NULL);
}


static void xmms68_play(char *filename)
{
  disk68_t *d = 0;

  audio_error = FALSE;

  /* Load disk */
  d = SC68app_load_diskfile(SC68os_absolute_fname(filename), 1);
  if (!d) {
    goto error;
  }

  /* Set as current application disk */
  if (SC68load(&app, d) <0) {
    goto error;
  }

  /* Set default track to first ... */
  if (SC68track(&app, 0)<0) {
    goto error;
  }

  /* Start current track */
  if (SC68play(&app) < 0) {
    goto error;
  }

  /* open XMMS audio stream */
  if (xmms68_ip.output->open_audio(FMT_S16_LE, 44100, 2) == 0) {
    audio_error = TRUE;
    goto error;
  }

  going = TRUE;
  xmms68_set_info(app.cur_disk, app.cur_track);
  pthread_create(&play_thread, NULL, play_loop, d);
  return;

error:
  if (d) {
    SC68_free(d);
  }
  going = FALSE;
  spool_error_message();
}

/** XMMS pause callback.
 *
 *    This function just call the output plugin pause function.
 */
static void xmms68_pause(short paused)
{
  xmms68_ip.output->pause(paused);
}

static int xmms68_get_time(void)
{
  int t, t2;
	if (audio_error)
		return -2;

  if (!going  && !xmms68_ip.output->buffer_playing()) {
    t2 = -1;
  }	else {
    t2 = xmms68_ip.output->output_time();
  }

	if (app.ready && app.play) {
	  t = app.time.elapsed_ms;
	}	else {
	  t = -1;
	}

  /* $$$ Can't return t else vis pluggin does not work correctly !! */
	return t2;
}

static void xmms68_song_info(char *filename, char **title, int *length)
{
  disk68_t *d;

  *length = -1;
  *title = 0;
  d = SC68app_load_diskfile(filename, 0);
  if (d) {
    gchar *name;

    name =
        g_strdup_printf("%s [%02u] / %s", d->name, d->nb_six,
                        d->mus[d->default_six].author);
    *title = name;
    *length = d->time * 1000;
    SC68_free(d);
  }
}

/** XMMS plugin description. */
static char xmms68_description[] = "Atari-ST and Amiga soundchip emulator " VERSION;

/* Called when the plugin is loaded
*/
static void xmms68_init(void)
{
  int r, frq;

  SC68os_pdebug("\nINIT XMMS68 plugin\n");

  /* check whether EMU68 is compiled without debug option */
  if (EMU68_sizeof_reg68() != sizeof(reg68_t)) {
    SC68error_add
        ("Internal : Bad EMU68 version : should not define SC68DEBUG ...");
    exit(3);
  }

  /* Set default configuration */
  set_sc68app_default(&app);

  /* Get exec path : for resource */
  if (SC68os_get_exec_path(xmms68_ip.filename) < 0) {
    exit(4);
  }

  /* Get the config file OR create if it doesn't exist */
  SC68os_pdebug("Read config file ...\n");
  if (SC68config_load(&app.config,
                      SC68os_resource_name(SC68rsc_config,
                                           SC68_DEFAULT_CONFIG_FILE)) < 0) {
    SC68error_get();
    SC68error_get();
    SC68config_save(&app.config,
                    SC68os_resource_name(SC68rsc_config,
                                         SC68_DEFAULT_CONFIG_FILE));
  }
  SC68os_pdebug("OK\n");

  /*SC68config_display(&app.config);*/

  /* Open application ( OS dependant code ) */
  r = SC68os_interface_open(xmms68_ip.filename, &app);

  /* Open error */
  if (r < 0) {
    return;
  }

  /* Copy config parameters to app */
  SC68app_configure(&app);

  /* Delay init : from config file parameters ( SC68app_configure() ) */
  app.delay.buffer_size = SC68delay_init(app.delay.buffer_size);
  if (app.delay.buffer_size < 0)
    return;

  /* Copy config parameters to app */
  SC68app_configure(&app);

  /* Delay init : from config file parameters ( SC68app_configure() ) */
  app.delay.buffer_size = SC68delay_init(app.delay.buffer_size);
  if (app.delay.buffer_size < 0)
    return;

  frq = (int) app.mix.real_frq;

  /* First instance : Do mucho initialization */
  if (r > 0) {
#ifdef WITH_BROWSE68
    /* Browse Info Data Base init */
    if (SC68browse_init(SC68_DEFAULT_BIDB_FILE) < 0) {
      return;
    }
    SC68os_pdebug("Browse init : OK\n");
#endif
    /* 68K & IO-Emulators init */
    if (init_68K(&app.reg68, MEMORY)) {
      spool_error_message();
      return;
    }
    SC68os_pdebug("68K emulator init : OK\n");

    /* IO replay frequency init */
    app.mix.real_frq = 44100;
    YM_set_replay_frq(app.mix.real_frq);
    MW_set_replay_frq((int) app.mix.real_frq);
    PL_set_replay_frq((int) app.mix.real_frq);
    app.mix.buf = YM_get_buffer();
  }

}

/** Clean exit XMMS callback. */
static void xmms68_cleanup(void)
{
  int i;

  SC68os_pdebug("\nCLEAN UP XMMS68 plugin\n");

  SC68os_interface_close(&app);
#ifdef WITH_BROWSE68
  SC68browse_write(SC68_DEFAULT_BIDB_FILE);
#endif
  if (i = SC68config_get_id("total_time"), i >= 0) {
    app.config[i] = app.time.total;
  }
  SC68config_save(&app.config,
                  SC68os_resource_name(SC68rsc_config, "config"));
  SC68os_pdebug("\nSC68 : Clean Exit\n");
  spool_error_message();
}

/** Seek XMMS callback.
 *
 *    This is used to change track in the current disk.
 */
static void xmms68_seek(int time)
{
  if (going && app.ready && app.cur_disk) {
    disk68_t *d = app.cur_disk;
    int i, cnt;

    for (i = cnt = 0; i < d->nb_six; ++i) {
      int cnt2;

      cnt2 = cnt + d->mus[i].time;
      if (time >= cnt && time < cnt2) {
        if (i != app.cur_track) {
          SC68track(&app, i);
          if (!audio_error && xmms68_ip.output) {
            xmms68_ip.output->flush (cnt * 1000);
          }
          xmms68_set_info(d, app.cur_track);
        } else {
          /*SC68os_pdebug("STAY AT %u\n", i + 1);*/
        }
      }
      cnt = cnt2;
    }
  }
}

/** XMMS input plugin instance. */
static InputPlugin xmms68_ip =
{
	NULL,                 /* !! Handle */
	NULL,                 /* !! Filename */
	xmms68_description,   /* Description */
	xmms68_init,          /* Init */
	xmms68_about,         /* Show about */
	NULL,                 /* Configure */
	xmms68_is_our_file,
	NULL,                 /* Scan dir */
	xmms68_play,
	xmms68_stop,
	xmms68_pause,
	xmms68_seek,          /* Seek */
	NULL,                 /* Set eq */
	xmms68_get_time,      /* */
	NULL,                 /* get volume l/r */
	NULL,                 /* set volume l/r */
	xmms68_cleanup,       /* Exit call */
	NULL,                 /* Get vis type (OBSOLETE) */
	NULL,                 /* Sends data to vis pluggin */
	NULL,                 /* !!! Set info */
	NULL,                 /* !!! Set song title */
	xmms68_song_info,
	NULL,                 /* file info box */
	NULL                  /* Output plugin */
};

/** Get input plugin info XMMS callback. */
InputPlugin *get_iplugin_info(void)
{
  return &xmms68_ip;
}
